/**
Master Project WS-2004 ~ SS-2005, Computer Science

University of Fribourg, Information Systems Research Group (http://diuf.unifr.ch/is)

Student :    Amos Brocco <amos.brocco@unifr.ch>
Supervisors: Prof. Dr. Andreas Meier <andreas.meier@unifr.ch>
             Nicolas Werro <nicolas.werro@unifr.ch>
        __  __
 _ _ ___\ \/ /__ _ _ _  _ _  ___
| '_/ _ \>  </ _` | ' \| ' \/ -_)
|_| \___/_/\_\__,_|_||_|_||_\___|

roXanne, Secure XML Data Storage Framework
x.core & x.click Copyright (C) 2004-2005 Amos Brocco <amos.brocco@unifr.ch>
x.mill Copyright (C) 2004-2005 Andrea Ghirlanda <andrea.ghirlanda@unifr.ch>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

This product includes software developed by the JDOM Project (http://www.jdom.org/).
The Jaxup library is Copyright (c) 2001 - 2003 Erwin Bolwidt, used and modified under the BSD License.
Encryption modules used are Copyright (c) 2000 - 2004 The Legion Of The Bouncy Castle (http://www.bouncycastle.org).
JAXEN  is Copyright 2003 (C) The Werken Company. All Rights Reserved.
Base64 was released as Public Domain by its author Robert Harder (rob@iharder.net)
MySQL JDBC Connector is (c) 1995-2005 MySQL AB. All rights reserved. Released under the GNU/GPL License.
PostgreSQL JDBC Driver is Copyright (c) 2004-2005 The PostgreSQL Global Development Group, distributed under the BSD license
JArgs was created Steve Purcell, now mantained by Ewan Mellorby and released under the BSD licence.
**/

/*
 * Environment
 * 
 * An environment holds a tree fragment. There exist a root environment
 * for XML data that is created when the database is started. Then
 * for every request that involves an ecrypted node, a new environment
 * gets created.
 */
package environments;

import java.util.Iterator;
import org.jdom.Document;
import org.jdom.Element;

/**
 * An environment is used to manage a DOM tree fragment. Each time an encrypted
 * node is visited we try to create an environment containing the DOM tree
 * constructed from the decrypted XML data. If the user has no key to decrypt
 * the node, the encrypted one is returned. If the user has ability only to
 * decrypt data, the environment and its elements are marked as being read only.
 * 
 * @author mriya3
 *  
 */
public class Environment {

	private String algorithm;

	private String algorithmName;

	private Element baseNode;

	private int deadEnv = 0;

	private EnvironmentManager envManager;

	private String keyName;

	private boolean modifiedflag = false;

	private Environment parentEnvironment;

	private Element rootNode;

	private boolean writable = true;

	/**
	 * Creates a new environment and link to the environment manager
	 * 
	 * @param envmgr
	 */
	public Environment(EnvironmentManager envmgr) {
		this.envManager = envmgr;
	}

	/**
	 * Returns the algorithm for this environment
	 * 
	 * @return
	 */
	public String getAlgorithm() {
		return this.algorithmName;
	}

	/**
	 * Returns the node that "generated" this environment.
	 * 
	 * @return
	 */
	public Element getBaseNode() {
		return this.baseNode;
	}

	public EnvironmentManager getEnvironmentManager() {
		return this.envManager;
	}

	/**
	 * Returns the key for this environment
	 * 
	 * @return
	 */
	public String getKeyName() {
		return this.keyName;
	}

	/**
	 * Returns the parent environment
	 * 
	 * @return
	 */
	public Environment getParent() {
		return this.parentEnvironment;
	}

	/**
	 * Constructs and returns a Document object that links to the root node of
	 * the environment. This is used to perform XPath and XUpdate queries which
	 * expect a document object instead of an element one.
	 * 
	 * @return
	 */
	public Document getRootDocument() {
		if (!(this.rootNode == null) && (this.rootNode.getParent() == null)) {
			Document doc = new Document();
			doc.addContent((Element) this.rootNode);
			return doc;
		} else if (!(this.rootNode == null)
				&& !(this == this.envManager.getRootEnvironment())) {
			return this.rootNode.getDocument();
		} else {
			return this.envManager.getRootEnvironment().getRootDocument();
		}
	}

	/**
	 * This method returns the root node of the environment, which is at the top
	 * of the environment XML tree.
	 * 
	 * @return
	 */
	public Element getRootNode() {
		return this.rootNode;
	}

	/**
	 * Returns the status of this environment
	 * 
	 * @return
	 */
	public int getStatus() {
		return this.deadEnv;
	}

	/**
	 * Returns true if the environment is alive (linked)
	 * 
	 * @return
	 */
	public boolean isAlive() {
		return (this.deadEnv == 1);
	}

	/**
	 * Returns true if the environment is dead (unlinked)
	 * 
	 * @return
	 */
	public boolean isDead() {
		return (this.deadEnv == -1);
	}

	/**
	 * Returns if the environment was modified
	 * 
	 * @return
	 */
	public boolean isModified() {
		return this.modifiedflag;
	}

	/**
	 * Returns if the environment is marked as readonly. The readonly property,
	 * accessed through this method is used by element modifiers via the
	 * getEnvironment method to check if the element is editable.
	 * 
	 * @return
	 */
	public boolean isReadonly() {
		return !(this.writable);
	}

	/**
	 * Returns true if the environment is in an unknown status
	 * 
	 * @return
	 */
	public boolean isUnknown() {
		return (this.deadEnv == 0);
	}

	/**
	 * Recursively sets the environment link in the node and in each child node.
	 * This method first switches off node visiting in order to avoid the
	 * creation of environments.
	 * 
	 * @param top
	 * @param env
	 */
	public void recursiveSetEnvironmentLink(Element top, Environment env) {
		top.setEnvironment(env);
		// Disable visiting child nodes to prevent the creation
		// of new environments when accessing them
		top.setVisitChildOff();
		Iterator childIterator = top.getChildren().iterator();
		while (childIterator.hasNext()) {
			Element child = (Element) childIterator.next();
			child.setEnvironment(env);
			child.getEnvironment().recursiveSetEnvironmentLink(child, env);
		}
		top.setVisitChildOn();
	}

	/**
	 * Sets the algorithm for this environment
	 * 
	 * @param algorithmName
	 */
	public void setAlgorithm(String algorithmName) {
		this.algorithmName = algorithmName;
	}

	/**
	 * Sets the environment as alive
	 *  
	 */
	public void setAlive() {
		this.deadEnv = 1;
	}

	/**
	 * Sets the generator node
	 * 
	 * @param base
	 */
	public void setBaseNode(Element base) {
		this.baseNode = base;
	}

	/**
	 * Sets the environment as dead
	 *  
	 */
	public void setDead() {
		this.envManager.setRootUnSynced();
		this.envManager.getRootEnvironment().setModified();
		this.deadEnv = -1;
	}

	public void setEnvironmentManager(EnvironmentManager em) {
		this.envManager = em;
	}

	/**
	 * Sets the key for this environment
	 * 
	 * @param keyName
	 */
	public void setKeyName(String keyName) {
		this.keyName = keyName;
	}

	/**
	 * Sets the environment as modified
	 *  
	 */
	public synchronized void setModified() {
		if ((!(this.envManager == null))
				&& (this.envManager.enabledModifyNotify())) {
			if (!(this.parentEnvironment == null))
				this.parentEnvironment.setModified();
			if ((this.parentEnvironment == envManager.getRootEnvironment()))
				((RootEnvironment) this.parentEnvironment).setUnSynced();
			if (this instanceof RootEnvironment)
				((RootEnvironment) this).setUnSynced();
			this.modifiedflag = true;
		}
	}

	/**
	 * Sets the parent environment
	 * 
	 * @param environment
	 */
	public void setParent(Environment env) {
		this.parentEnvironment = env;
	}

	/**
	 * Sets the environment as read only, meaning that no modification is
	 * possible for elements contained in this environment and in derived ones.
	 *  
	 */
	public void setReadonly() {
		this.writable = false;
	}

	/**
	 * Sets a link to the root element
	 * 
	 * @param root
	 */
	public void setRootNode(Element root) {
		this.rootNode = root;
	}

	/**
	 * Sets the status as unknown, meaning that this environment should be
	 * checked again
	 *  
	 */
	public void setUnknown() {
		this.deadEnv = 0;
	}

	/**
	 * Sets the environment as not modified
	 *  
	 */
	public synchronized void setUnmodified() {
		this.modifiedflag = false;
	}

	/**
	 * Sets the environment as writable
	 */
	public void setWritable() {
		this.writable = true;
	}

	/**
	 * Visiting a node means returning the node itself if it's not an encrypted
	 * one or creating an environment for XML plain data and returning a pointer
	 * to the root node of the environment.
	 * 
	 * @param theElement
	 * @return
	 */
	public Element visitNode(Element theElement) {
		if ((this.envManager == null)
				|| !(this.envManager.isVisitNodeEnabled())
				|| (!theElement.isEncryptedNode()))
			return theElement;

		// Gets or Creates a new environment and get its root node
		return this.envManager.setEnvironment(theElement);
	}

}
