package org.jdom.contrib.c14n;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import org.jdom.*;
import org.jdom.filter.Filter;

/**
 * <p>
 * This class produces XML as specified by <a
 * href="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"> Canonical XML Version
 * 1.0 </a>
 * </p>
 * 
 * <p>
 * Currently, only complete JDOM Documents are supported.
 * </p>
 * 
 * <p>
 * Note that <code>Document</code> s that contain <code>EntityRef</code>s,
 * cannot be canonicalized, since the canonicalization algorithm prescribes
 * entity expansion. For example, trying to canonicalize a document that was
 * built with SAXBuilder.setExpandEntities(false), may yield a RuntimeException.
 * </p>
 * 
 * <p>
 * Credits are due to Elliotte Rusty Harold for the Canonicalizer from his <a
 * href="http://www.cafeconleche.org/XOM/">XOM </a> software
 * </p>
 * 
 * @author <a href="mailto:per.norrman@austers.se">Per Norrman </a>
 *  
 */
public class Canonicalizer {

	private Writer _out;

	private boolean _withComments = false;

	private Comparator _attributeComparator = new AttributeComparator();

	private Comparator _namespaceComparator = new NamespaceComparator();

	private NamespaceStack _namespaceStack;

	private List _namespaceList = new ArrayList();

	/**
	 * Creates a Canonicalizer.
	 * 
	 * @param out
	 *            The <code>OutputStream</code> on to which the document will
	 *            be written
	 * 
	 * @param withComments
	 *            true if comments should bew included in the output, otherwise
	 *            false
	 */
	public Canonicalizer(OutputStream out, boolean withComments) {
		_withComments = withComments;
		try {
			_out = new BufferedWriter(new OutputStreamWriter(out, "UTF8"));
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(
					"Severe error! Doest not understand UTF-8!");
		}
	}

	/**
	 * Creates a Canonicalizer that includes comments.
	 * 
	 * @param out
	 *            The <code>OutputStream</code> on to which the document will
	 *            be written
	 */
	public Canonicalizer(OutputStream out) {
		this(out, false);
	}

	/**
	 * Write the canonical XML for a <code>Document</code> on the
	 * <code>OutputStream</code> that was specified for this Canonicalizer.
	 * 
	 * @param doc
	 * 
	 * @throws IOException
	 *             if there is an I/O error on the underlying output stream
	 * 
	 * @throws RuntimeException
	 *             if the Document contains at least one <code>EntityRef</code>
	 *             instance.
	 *  
	 */
	public final void write(Document doc) throws IOException {
		_namespaceStack = new NamespaceStack();
		List l = doc.getContent(new Filter() {
			/**
			 * Comment for <code>serialVersionUID</code>
			 */
			private static final long serialVersionUID = -8222444192255444167L;

			public boolean matches(Object obj) {
				return obj instanceof ProcessingInstruction
						|| obj instanceof Comment || obj instanceof Element;
			}
		});

		int index = 0;
		while (true) {
			Content child = (Content) l.get(index);
			writeContent(child);
			index++;
			if (child instanceof ProcessingInstruction) {
				_out.write('\n');
			} else if (child instanceof Comment && _withComments) {
				_out.write('\n');
			} else if (child instanceof Element)
				break;
		}

		for (int i = index; i < l.size(); i++) {
			Content child = (Content) l.get(i);
			if (child instanceof ProcessingInstruction) {
				_out.write('\n');
			} else if (child instanceof Comment && _withComments) {
				_out.write('\n');
			}
			writeContent(child);
		}

		_out.flush();
	}

	private void writeContent(Content c) throws IOException {
		if (c instanceof Element) {
			write((Element) c);
		} else if (c instanceof Text) {
			write((Text) c);
		} else if (c instanceof ProcessingInstruction) {
			write((ProcessingInstruction) c);
		} else if (c instanceof Comment) {
			write((Comment) c);
		} else if (c instanceof EntityRef) {
			throw new RuntimeException(
					"Cannot produce canonical XML from unexpanded entities");
		}
	}

	private void write(ProcessingInstruction pi) throws IOException {
		_out.write("<?");
		_out.write(pi.getTarget());
		if (pi.getData().length() > 0) {
			_out.write(' ');
			_out.write(pi.getData());
		}
		_out.write("?>");
	}

	private void write(Comment comment) throws IOException {
		if (_withComments) {
			_out.write("<!--");
			_out.write(comment.getText());
			_out.write("-->");
		}
	}

	private void write(Element e) throws IOException {
		int namespaceMark = _namespaceStack.size();
		writeStartTag(e);
		writeEndTag(e);
		// remove declared namespaces from stack
		while (_namespaceStack.size() > namespaceMark) {
			_namespaceStack.pop();
		}
	}

	private void writeStartTag(Element e) throws IOException {
		_out.write("<");
		_out.write(e.getQualifiedName());
		writeNamespaces(e);
		writeAttributes(e);
		_out.write(">");
		for (Iterator i = e.getContent().iterator(); i.hasNext();) {
			writeContent((Content) i.next());
		}
	}

	private void writeNamespaces(Element e) throws IOException {
		_namespaceList.clear();

		addElementNamespace(e);
		for (Iterator i = e.getAdditionalNamespaces().iterator(); i.hasNext();) {
			addNamespace((Namespace) i.next());
		}

		// add attribute namespaces but not for the empty namespace or XML
		// namespace
		for (Iterator i = e.getAttributes().iterator(); i.hasNext();) {
			Attribute a = (Attribute) i.next();
			Namespace ns = a.getNamespace();
			if ((ns != Namespace.NO_NAMESPACE)
					&& (ns != Namespace.XML_NAMESPACE)) {
				addNamespace(ns);
			}
		}

		/*
		 * The _namespaceList now contains the list of namespaces that we should
		 * print, but they must first be sorted
		 */
		Namespace[] nsa = (Namespace[]) _namespaceList
				.toArray(new Namespace[0]);
		Arrays.sort(nsa, _namespaceComparator);
		for (int i = 0; i < nsa.length; i++) {
			String prefix = nsa[i].getPrefix();
			String uri = nsa[i].getURI();
			_out.write(" xmlns");
			if (!prefix.equals("")) {
				_out.write(":");
				_out.write(prefix);
			}
			_out.write("=\"");
			_out.write(uri);
			_out.write("\"");
		}

	}

	private void addElementNamespace(Element element) {
		// Add namespace decl only if it's not the XML namespace and it's
		// not the NO_NAMESPACE with the prefix "" not yet mapped
		// (we do output xmlns="" if the "" prefix was already used and we
		// need to reclaim it for the NO_NAMESPACE)
		Namespace ns = element.getNamespace();
		if (ns == Namespace.XML_NAMESPACE) {
			return;
		}
		if (!((ns == Namespace.NO_NAMESPACE) && (_namespaceStack.getURI("") == null))) {
			addNamespace(ns);
		}
	}

	private void addNamespace(Namespace ns) {
		String prefix = ns.getPrefix();
		String uri = ns.getURI();

		// Already added namespace decl?
		if (uri.equals(_namespaceStack.getURI(prefix))) {
			return;
		}

		_namespaceList.add(ns);
		_namespaceStack.push(ns);
	}

	private void writeAttributes(Element e) throws IOException {
		List l = e.getAttributes();
		if (l.size() > 0) {
			Attribute[] attributes = new Attribute[l.size()];
			for (int i = 0; i < attributes.length; i++) {
				attributes[i] = (Attribute) l.get(i);
			}
			Arrays.sort(attributes, _attributeComparator);
			for (int i = 0; i < attributes.length; i++) {
				write(attributes[i]);
			}
		}
	}

	private void write(Attribute attribute) throws IOException {
		_out.write(' ');
		_out.write(attribute.getQualifiedName());
		_out.write("=\"");
		String s = attribute.getValue();
		for (int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			switch (c) {
			case '&':
				_out.write("&amp;");
				break;
			case '<':
				_out.write("&lt;");
				break;
			case '"':
				_out.write("&quot;");
				break;
			case '\t':
				_out.write("&#x9;");
				break;
			case '\n':
				_out.write("&#xA;");
				break;
			case '\r':
				_out.write("&#xD;");
				break;
			default:
				_out.write(c);
			}
		}
		_out.write("\"");
	}

	private void writeEndTag(Element e) throws IOException {
		_out.write("</");
		_out.write(e.getQualifiedName());
		_out.write(">");
	}

	private void write(Text text) throws IOException {
		String s = text.getText();
		int l = s.length();
		for (int i = 0; i < l; i++) {
			char c = s.charAt(i);
			switch (c) {
			case '&':
				_out.write("&amp;");
				break;
			case '<':
				_out.write("&lt;");
				break;
			case '>':
				_out.write("&gt;");
				break;
			case '\r':
				_out.write("&#xD;");
				break;
			default:
				_out.write(c);
			}
		}
	}

}
