/**
Master Project WS-2004 ~ SS-2005, Computer Science

University of Fribourg, Information Systems Research Group (http://diuf.unifr.ch/is)

Student :    Amos Brocco <amos.brocco@unifr.ch>
Supervisors: Prof. Dr. Andreas Meier <andreas.meier@unifr.ch>
             Nicolas Werro <nicolas.werro@unifr.ch>
        __  __
 _ _ ___\ \/ /__ _ _ _  _ _  ___
| '_/ _ \>  </ _` | ' \| ' \/ -_)
|_| \___/_/\_\__,_|_||_|_||_\___|

roXanne, Secure XML Data Storage Framework
x.core & x.click Copyright (C) 2004-2005 Amos Brocco <amos.brocco@unifr.ch>
x.mill Copyright (C) 2004-2005 Andrea Ghirlanda <andrea.ghirlanda@unifr.ch>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

This product includes software developed by the JDOM Project (http://www.jdom.org/).
The Jaxup library is Copyright (c) 2001 - 2003 Erwin Bolwidt, used and modified under the BSD License.
Encryption modules used are Copyright (c) 2000 - 2004 The Legion Of The Bouncy Castle (http://www.bouncycastle.org).
JAXEN  is Copyright 2003 (C) The Werken Company. All Rights Reserved.
Base64 was released as Public Domain by its author Robert Harder (rob@iharder.net)
MySQL JDBC Connector is (c) 1995-2005 MySQL AB. All rights reserved. Released under the GNU/GPL License.
PostgreSQL JDBC Driver is Copyright (c) 2004-2005 The PostgreSQL Global Development Group, distributed under the BSD license
JArgs was created Steve Purcell, now mantained by Ewan Mellorby and released under the BSD licence.
**/

/*
 * Session management
 */
package sessions;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.Semaphore;

import org.jdom.output.XMLOutputter;

import server.MultithreadedServer;

import application.ConfigurationRegistry;

import debug.Debuginfo;

import environments.DataSynchronizer;
import environments.Environment;
import environments.EnvironmentUpdater;
import environments.RootEnvironment;

/**
 * Session Manager There is only one session manager object. The session manager
 * also provides a global semaphore for accessing the XML structure.
 * 
 * @author mriya3
 */
public class SessionManager {

	private ConfigurationRegistry conf;

	public final Semaphore domAccess;

	private EnvironmentUpdater envUpdater;
	
	private DataSynchronizer dataSync;

	private Session lastPerformer;

	private XMLOutputter op;

	private RootEnvironment rootenv;

	Hashtable<String, Session> sessions;

	private Vector<Environment> updaterlist;

	private MultithreadedServer server;

	public SessionManager(ConfigurationRegistry confreg, RootEnvironment rt) {
		this.conf = confreg;
		domAccess = new Semaphore(1, true);
		envUpdater = new EnvironmentUpdater(this);
		dataSync = new DataSynchronizer(this);
		this.rootenv = rt;
		this.sessions = new Hashtable<String, Session>();
		op = new XMLOutputter();
		envUpdater.start();
		dataSync.start();
	}

	/**
	 * Called by someone that acquired the lock on the data tree
	 * 
	 * @param s
	 */
	public void acquiredDataLock(Session s) {
			/*
			 * Stop the updater - This is a blocking call meaning that we can
			 * continue only if the updater as really stopped
			 * stopUpdater();
			 */
			/* Check if a complete update is necessary */
			flushUpdatesIfNecessary(s);
			rootenv.setEnvironmentManager(s.getEnvironmentManager());
			/* If the root is synced with the rest of the tree save to disk */
			s.getEnvironmentManager().rootIsUpToDate();
			if(!(s.getRootEnvironment().getSyncedStatus()) && (s.getRootEnvironment().isModified())) {
				saveUpdates();
			}
	}

	/**
	 * Checks login name and password
	 * 
	 * @param login
	 * @param password
	 * @return
	 */
	public boolean checkLogin(String login, String password) {
		return conf.getStringValue("login(" + login + ")", "!" + password)
				.equals(password);
	}

	/**
	 * Creates a new session
	 * 
	 * @return
	 */
	public synchronized String createSession(String login, String address) {
		String id = null;
		Session newsession = new Session(this);
		id = IDGenerator.getRandomID();
		// First purge unneeded old sessions
		purgeSessions();
		// Just to be sure not to generate two same id's
		while (sessions.contains(login + "@" + address + "@" + id))
			id = IDGenerator.getRandomID();
		// To retrieve a session we use the key IP@sessionid
		sessions.put(new String(login + "@" + address + "@" + id), newsession);
		newsession.setRootEnvironment(this.rootenv);
		newsession.updateTimestamp();
		this.rootenv.setEnvironmentManager(newsession.getEnvironmentManager());
		return id;
	}

	/**
	 * Deletes a session
	 * 
	 * @param sessionID
	 */
	public synchronized void deleteSession(String login, String address,
			String sessionID) {
		if (sessions.containsKey(new String(login + "@" + address + "@" + sessionID))) {
			Session ss = (Session) sessions.get(login + "@" + address + "@" + sessionID);
			try {
				domAccess.acquire();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			ss.getEnvironmentManager().closeAllEnvironments();
			domAccess.release();
			sessions.remove(new String(login + "@" + address + "@" + sessionID));
		}
	}

	/**
	 * Returns true if the session exists
	 * 
	 * @param reqSid
	 * @return
	 */
	public boolean existSession(String login, String address, String sessionID) {
		return sessions.containsKey(new String(login + "@" + address + "@" + sessionID));
	}

	/**
	 * This is a little optimization trick: if the Session object (here called
	 * performer) is not the same as the lastperformer, then force closing all
	 * session environments. If they are the same, there is no reason to close
	 * environments, as the caller is the same :) After flushing all data, XML
	 * is saved on disk
	 * 
	 * @param performer
	 */
	public synchronized void  flushUpdatesIfNecessary(Session performer) {
		if (!(lastPerformer == null))
			if (performer != lastPerformer) {
				lastPerformer.getEnvironmentManager().closeAllEnvironments();
			}
		lastPerformer = performer;
	}

	/**
	 * Returns the configuration registry object
	 * 
	 * @return
	 */
	public ConfigurationRegistry getRegistry() {
		return conf;
	}

	/**
	 * Returns a session. If the session is too old, close the session and
	 * return null;
	 * 
	 * @param sessionID
	 * @return
	 */
	public synchronized Session getSession(String login, String address,
			String sessionID) {
		if (sessions.containsKey(login + "@" + address + "@" + sessionID)) {
			Session tmps = sessions
					.get(login + "@" + address + "@" + sessionID);
			tmps.getRootEnvironment().setEnvironmentManager(
					tmps.getEnvironmentManager());
			return tmps;

		} else
			return null;
	}

	/**
	 * Returns the number of currently opened sessions
	 * 
	 * @return
	 */
	public int getSessionsCount() {
		return sessions.size();
	}

	/**
	 * Checks if the login name is an administrator
	 * 
	 * @param login
	 * @return
	 */
	public boolean isAdministrator(String login) {
		return conf.getStringValue("role(" + login + ")", "!administrator")
				.equals("administrator");
	}

	/**
	 * Removes old sessions from the list
	 *  
	 */
	public synchronized void purgeSessions() {
		Date date = new Date();
		long currentTicks = date.getTime();
		Object[] slist = (Object[]) sessions.keySet().toArray();
		for(int v=0; v < slist.length; v++ ) {
			String skey = (String) slist[v];
			Session tmps = (Session)sessions.get(skey);
			if (((currentTicks - tmps.getTimestamp()) > conf.getLongValue(
					"SessionMaxTime", 1800000))
					|| !tmps.persistent)
				if (!tmps.equals(lastPerformer))
					sessions.remove(skey);
				else if (!lastPerformer.persistent)
					sessions.remove(skey);
		}
	}

	/**
	 * Called when the lock on the data tree is released
	 * 
	 * @param s
	 */
	public void releasedDataLock(Session s) {
	}
	
	
	
	/**
	 * Called when the lock on the data tree is about to be released
	 * 
	 * @param s
	 */
	public void aboutToReleaseDataLock(Session s) {
		s.getEnvironmentManager().setDeadOrAlive();
		s.getEnvironmentManager().deleteAllDeadEnvironments();
		setUpdaterList(s.getEnvironmentManager().getEnvironmentList());
	}
	
	

	/**
	 * Saves updates (the root structure) to disk
	 * 
	 * This method saves the root environment to disk. As it is launched when
	 * all environments are either closed or correctly updated we ensure that
	 * all data is up to date.
	 *  
	 */
	public void saveUpdates() {
		/* Disable node visiting */
		if (!(lastPerformer == null))
			lastPerformer.getEnvironmentManager().disableNodeVisiting();
		/* Set bogus env manager */
		/* Save XML data to disk */
		BufferedWriter outfile = null;
		String xmlrootfile = conf.getStringValue("XMLdbfile", "root.xml");
		try {
			outfile = new BufferedWriter(new FileWriter(xmlrootfile));
		} catch (IOException e1) {
			new Debuginfo(4,
					"Cannot open output file to flush updates to disk: "
							+ e1.getLocalizedMessage());
		}
		try {
				op.output(rootenv.getRootDocument(), outfile);
		} catch (IOException e2) {
			new Debuginfo(4, "Cannot write to outputfile: "
					+ e2.getLocalizedMessage());
		} finally {
			try {
				outfile.flush();
				outfile.close();
				this.rootenv.setUnmodified();
				this.rootenv.setSynced();
			} catch (IOException e) {
				new Debuginfo(4, "Cannot close output file: "
						+ e.getLocalizedMessage());
			}
		}
		/* Re-enable node visiting */
		if (!(lastPerformer == null))
			lastPerformer.getEnvironmentManager().enableNodeVisiting();

	}

	/**
	 * Sets the list of environments to be updated for the updater thread
	 * 
	 * @param updatelist
	 */
	public synchronized void setUpdaterList(Vector<Environment> updatelist) {
		this.updaterlist = (Vector<Environment>) updatelist.clone();
	}
	
	public synchronized Vector<Environment> getUpdaterList() {
		return this.updaterlist;
	}

	/**
	 * Starts the environment updater thread
	 *  
	 */
	public synchronized void startUpdater() {
		envUpdater.resume();
	}

	/**
	 * Stops the environment updater thread
	 */
	public synchronized void stopUpdater() {
		envUpdater.interrupt();
	}

	/**
	 * Updates data to disk when requesting application termination
	 */
	public void onClosing() {
		if (!(lastPerformer == null)) 
			lastPerformer.getEnvironmentManager().closeAllEnvironments();
		saveUpdates();
		new Debuginfo(1,"XML data flushed to disk");
	}
	
	/**
	 * Returns the last session that executed something, i.e.
	 * acquired the data lock 
	 * 
	 * @return
	 */
	public Session getLastPerformer() {
		return lastPerformer;
	}

	/**
	 * Returns the root environment
	 * @return
	 */
	public RootEnvironment getRootEnvironment() {
		return this.rootenv;
	}

	/**
	 * Sets the server link
	 * @param server
	 */
	public void setServer(MultithreadedServer server) {
		this.server = server;
	}
	
	/**
	 * Returns the associated Multithreadedserver
	 * @return
	 */
	public MultithreadedServer getServer() {
		return this.server;
	}



}
